//-------------------------------------------------------------------
// Prefetch Speed Patch
//-------------------------------------------------------------------


#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <io.h>


// Speed Patcher version
//
const char *version = "Prefetch Speed Patcher v0.9";


// gameboy rom signature
//
unsigned char gbaromsig[] =
{
    0x24, 0xff, 0xae, 0x51, 0x69, 0x9a, 0xa2, 0x21, 
    0x3d, 0x84, 0x82, 0x0a, 0x84, 0xe4, 0x09, 0xad,
};


// activation code for prefetch state
// (this is an ARM-mode code that will be added to the end of the ROM file)
//
unsigned char code[] =
{
    // 32 bytes of GBA ARM asm code is all it takes to do the trick.
    // Generated with Goldroad assembler.
    // 
    // ld r0, =$04000204        ; WAITCNT waitstate control register
    // ld r1, =$00004000        ; turns on prefetch, sets every thing else to default
    // str r1, [r0]
    // ld r0, =$00000000        ; this will store the return addr
    // mov r15, r0              ; returns back to the original entry address.
    //
    0x0c, 0x00, 0x9f, 0xe5, 0x0c, 0x10, 0x9f, 0xe5, 
    0x00, 0x10, 0xa0, 0xe4, 0x08, 0x00, 0x9f, 0xe5,
    0x00, 0xf0, 0xa0, 0xe1, 0x04, 0x02, 0x00, 0x04,
    0x02, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 
};


// space to load bytes for testing
//
char space[4096];


//-------------------------------------------------------------------
// Determine if the patch is contained in the file at the
// location.
//-------------------------------------------------------------------
bool containspatch( FILE *fp, int loc )
{
    fseek( fp, loc, SEEK_SET );

    // search for the patcher signature
    //
    int patchcodelength = sizeof(code);
    bool diff = false;
    for( int i=0; i<patchcodelength-4; i++ )
    {
        int c = fgetc( fp );
        if( c!=code[i] )
        {
            diff = true;
            break;
        }
    }
    return !diff;
}


//-------------------------------------------------------------------
// Find unused block within the code.
//-------------------------------------------------------------------
int findunusedblock( char *filename, FILE *fp, int blocksize )
{
    printf( "%s...\r", filename );

    if( blocksize>=sizeof(space) )
        blocksize = sizeof(space);

    fseek( fp, 0, SEEK_END );
    int filesize = ftell( fp );
    int loc = ((filesize - blocksize) / blocksize) * blocksize;

    while( loc >= 0 )
    {
        printf( "%s: %08X\r", filename, loc );

        fseek( fp, loc, SEEK_SET );
        fgets( space, blocksize, fp );

        if( containspatch( fp, loc + blocksize/2 ) )
            return -(loc + blocksize/2);

        bool unused = true;
        if( space[0]==0 )
        {
            for( int i=0; i<blocksize; i++ )
                if( space[0]!=0 )
                {
                    unused = false;
                    break;
                }
        }
        else if( space[0]==0xFF )
        {
            bool unused = true;
            for( int i=0; i<blocksize; i++ )
                if( space[0]!=0xFF )
                {
                    unused = false;
                    break;
                }
        }

        if( unused )
            return loc + blocksize/2;

        loc -= blocksize;
    }
    return 0;
}


//-------------------------------------------------------------------
// Patch file
//-------------------------------------------------------------------
void patchfile( char *filename )
{
    // get the short file name
    //
    char shortfilename[41];
    int filenamestart = 0;
    for( int j=0; j<sizeof(shortfilename); j++ )
        shortfilename[j] = 0;
    for( int j=0; j<255 && filename[j]!=0; j++ )
        if( filename[j]=='\\' )
            filenamestart = j+1;
    for( int j=0; j<sizeof(shortfilename) && filename[j+filenamestart]!=0; j++ )
        shortfilename[j] = filename[j+filenamestart];
    shortfilename[sizeof(shortfilename)-1] = 0;

    // test if the file exists or if it can be opened for write.
    // 
    FILE *fp = fopen( filename, "r+b" );
    if( fp==NULL )
    {
        printf( "%-40s: File does not exist.\n", shortfilename );
        return;
    }

    fseek( fp, 0, SEEK_END );
    int filesize = ftell( fp );

    // test the gameboy rom signature
    //
    bool gbarom = true;
    fseek( fp, 4, SEEK_SET );
    for( int i=0; i<sizeof(gbaromsig); i++ )
    {
        int c = fgetc( fp );
        if( c!=gbaromsig[i] )
        {
            gbarom = false;
            break;
        }
    }
    if( !gbarom )
    {
        printf( "%-40s: Not a gameboy ROM.\n", shortfilename );
        return;
    }

    // search for the patcher signature at the end of the file
    //
    int patchcodelength = sizeof(code);
    fseek( fp, filesize - patchcodelength, SEEK_SET );
    bool diff = false;
    for( int i=0; i<patchcodelength-4; i++ )
    {
        int c = fgetc( fp );
        if( c!=code[i] )
        {
            diff = true;
            break;
        }
    }
    if( !diff )
    {
        printf( "%-40s: Patched previously at $%08X.\n", shortfilename, filesize - patchcodelength );
        return;
    }

    // determine entry point for our patch code
    //
    int newentrypoint = filesize;
    if( filesize>=32768*1024-32 )
    {
        newentrypoint = findunusedblock( shortfilename, fp, 4096 );
        if( newentrypoint<0 )
        {
            printf( "%-40s: Patched previously at $%08X\n", shortfilename, -newentrypoint );
            return;
        }
        if( newentrypoint==0 )
        {
            printf( "%-40s: Can't find unused space.\n", shortfilename );
            return;
        }
    }
    int patchentrypoint = newentrypoint;
    
    // obtain the original entry point of the ROM, 
    // and insert it into the patch code
    //
    fseek( fp, 0, SEEK_SET );
    int entrypoint = (fgetc( fp ) + fgetc( fp ) * 256 + fgetc( fp ) * 65536)*4+8 + 0x8000000;

    code[patchcodelength-4] = (entrypoint & 0xFF);
    code[patchcodelength-3] = ((entrypoint>>8) & 0xFF);
    code[patchcodelength-2] = ((entrypoint>>16) & 0xFF);
    code[patchcodelength-1] = ((entrypoint>>24) & 0xFF);

    // overwrite the first entry point 0x8000000 to force GBA to branch to 
    // our patch code.
    //
    int branchoffset = (patchentrypoint-8) / 4;
    fseek( fp, 0, SEEK_SET );
    fputc( branchoffset & 0xFF, fp );
    fputc( (branchoffset>>8) & 0xFF, fp );
    fputc( (branchoffset>>16) & 0xFF, fp );

    // now append our patch code into the end of the ROM file,
    // or overwrite at the patch address.
    //
    fseek( fp, patchentrypoint, SEEK_SET );
    for( int i=0; i<patchcodelength; i++ )
        fputc( code[i], fp );

    fclose( fp );
    printf( "%-40s: PATCHED SUCCESSFULLY at $%08X.\n", shortfilename, patchentrypoint );
}


//-------------------------------------------------------------------
// Main
//-------------------------------------------------------------------
void main(int argc, char **argv)
{
    printf( "%s\n", version );
    printf( "\n" );
    if( argc==1 )
    {
        printf( "Usage:\n" );
        printf( "  pfpatch rom1.gba rom2.gba rom3.gba ...\n" );
        printf( "  pfpatch dir1 dir2 dir3 ...\n" );
        printf( "\n" );
        printf( "Hint:\n" );
        printf( "  1. ROMs must have been patched with 'Enable Saver' (QPC Save) only.\n" );
        printf( "  2. Drag the ROM files or the folder onto the pfpatch.exe file.\n" );
        printf( "     (The number of ROM files you can drag-drop is limited to about 20-30),.\n" );
        printf( "\n" );
        printf( "The patch overwrites the ROM and does not keep a backup of the previous ROM. Please use at your own risk.\n" );
        return;
    }

    if( argc>1 )
    {
        
        for( int i=1; i<argc; i++ )
        {
            _finddata_t cfile;
            int h = _findfirst( argv[i], &cfile );
            if( h==-1 )
            {
                printf( "%s: Invalid file specification\n", argv[i] );
                continue;
            }

            if( cfile.attrib & _A_SUBDIR )
            {
                char dirpatch[1024];
                strncpy( dirpatch, argv[i], sizeof(dirpatch) );
                strncat( dirpatch, "\\*.gba", sizeof(dirpatch) );
                printf( "%s: searching subdirectory...\n", dirpatch );

                int h2 = _findfirst( dirpatch, &cfile );
                int err = h2;
                while( err!=-1 )
                {
                    if( !(cfile.attrib & _A_SUBDIR) )
                    {
                        char filepath[1024];
                        strncpy( filepath, argv[i], sizeof(filepath) );
                        strncat( filepath, "\\", sizeof(filepath) );
                        strncat( filepath, cfile.name, sizeof(filepath) );
                        patchfile( filepath );
                    }
                    err = _findnext( h2, &cfile );
                }
                _findclose( h2 );
            }
            else
            {
                patchfile( argv[i] );
            }
            _findclose( h );
        }
        printf( "Press any key...\n" );
        getch();
    }
}

